import logging
from typing import List, Optional

from firebolt.common.exception import AlreadyBoundError
from firebolt.common.urls import (
    ACCOUNT_BINDINGS_URL,
    ACCOUNT_DATABASE_BINDING_URL,
)
from firebolt.common.util import prune_dict
from firebolt.model.binding import Binding, BindingKey
from firebolt.model.database import Database
from firebolt.model.engine import Engine
from firebolt.service.base import BaseService

logger = logging.getLogger(__name__)


class BindingService(BaseService):
    def get_by_key(self, binding_key: BindingKey) -> Binding:
        """Get a binding by it's BindingKey"""
        response = self.client.get(
            url=ACCOUNT_DATABASE_BINDING_URL.format(
                account_id=binding_key.account_id,
                database_id=binding_key.database_id,
                engine_id=binding_key.engine_id,
            )
        )
        binding: dict = response.json()["binding"]
        return Binding.parse_obj(binding)

    def get_many(
        self,
        database_id: Optional[str] = None,
        engine_id: Optional[str] = None,
        is_system_database: Optional[bool] = None,
    ) -> List[Binding]:
        """
        List bindings on Firebolt, optionally filtering by database and engine.

        Args:
            database_id:
                Return bindings matching the database_id.
                If None, match any databases.
            engine_id:
                Return bindings matching the engine_id.
                If None, match any engines.
            is_system_database:
                If True, return only system databases.
                If False, return only non-system databases.
                If None, do not filter on this parameter.

        Returns:
            List of bindings matching the filter parameters.
        """

        response = self.client.get(
            url=ACCOUNT_BINDINGS_URL.format(account_id=self.account_id),
            params=prune_dict(
                {
                    "page.first": 5000,  # FUTURE: pagination support w/ generator
                    "filter.id_database_id_eq": database_id,
                    "filter.id_engine_id_eq": engine_id,
                    "filter.is_system_database_eq": is_system_database,
                }
            ),
        )
        return [Binding.parse_obj(i["node"]) for i in response.json()["edges"]]

    def get_database_bound_to_engine(self, engine: Engine) -> Optional[Database]:
        """Get the Database to which an engine is bound, if any."""
        try:
            binding = self.get_many(engine_id=engine.engine_id)[0]
        except IndexError:
            return None
        try:
            return self.resource_manager.databases.get(id_=binding.database_id)
        except (KeyError, IndexError):
            return None

    def get_engines_bound_to_database(self, database: Database) -> List[Engine]:
        """Get a list of engines that are bound to a database."""

        bindings = self.get_many(database_id=database.database_id)
        return self.resource_manager.engines.get_by_ids(
            ids=[b.engine_id for b in bindings]
        )

    def create(
        self, engine: Engine, database: Database, is_default_engine: bool
    ) -> Binding:
        """
        Create a new binding between an engine and a database.

        Args:
            engine: Engine to bind.
            database: Database to bind.
            is_default_engine:
                Whether this engine should be used as default for this database.
                Only one engine can be set as default for a single database.
                This will overwrite any existing default.

        Returns:
            New binding between the engine and database.
        """

        existing_database = self.get_database_bound_to_engine(engine=engine)
        if existing_database is not None:
            raise AlreadyBoundError(
                f"The engine {engine.name} is already bound "
                f"to {existing_database.name}!"
            )

        logger.info(
            f"Attaching Engine (engine_id={engine.engine_id}, name={engine.name}) "
            f"to Database (database_id={database.database_id}, "
            f"name={database.name})"
        )
        binding = Binding(
            binding_key=BindingKey(
                account_id=self.account_id,
                database_id=database.database_id,
                engine_id=engine.engine_id,
            ),
            is_default_engine=is_default_engine,
        )

        response = self.client.post(
            url=ACCOUNT_DATABASE_BINDING_URL.format(
                account_id=self.account_id,
                database_id=database.database_id,
                engine_id=engine.engine_id,
            ),
            json=binding.jsonable_dict(
                by_alias=True, include={"binding_key": ..., "is_default_engine": ...}
            ),
        )
        return Binding.parse_obj(response.json()["binding"])



from __future__ import annotations

from functools import wraps
from threading import Lock
from typing import Any, Dict, Generator, List, Optional, Sequence, Tuple

from readerwriterlock.rwlock import RWLockWrite

from firebolt.async_db._types import ColType
from firebolt.async_db.cursor import BaseCursor as AsyncBaseCursor
from firebolt.async_db.cursor import (
    ParameterType,
    check_not_closed,
    check_query_executed,
)
from firebolt.common.util import AsyncJobThread, async_to_sync


class Cursor(AsyncBaseCursor):
    """
    Class, responsible for executing queries to Firebolt Database.
    Should not be created directly,
    use :py:func:`connection.cursor <firebolt.async_db.connection.Connection>`

    Args:
        description: Information about a single result row
        rowcount: The number of rows produced by last query
        closed: True if connection is closed, False otherwise
        arraysize: Read/Write, specifies the number of rows to fetch at a time
            with the :py:func:`fetchmany` method
    """

    __slots__ = AsyncBaseCursor.__slots__ + (
        "_query_lock",
        "_idx_lock",
        "_async_job_thread",
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._query_lock = RWLockWrite()
        self._idx_lock = Lock()
        self._async_job_thread: AsyncJobThread = kwargs.pop("async_job_thread")
        super().__init__(*args, **kwargs)

    @wraps(AsyncBaseCursor.execute)
    def execute(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        set_parameters: Optional[Dict] = None,
        skip_parsing: bool = False,
    ) -> int:
        with self._query_lock.gen_wlock():
            return async_to_sync(super().execute, self._async_job_thread)(
                query, parameters, set_parameters, skip_parsing
            )

    @wraps(AsyncBaseCursor.executemany)
    def executemany(
        self, query: str, parameters_seq: Sequence[Sequence[ParameterType]]
    ) -> int:
        with self._query_lock.gen_wlock():
            return async_to_sync(super().executemany, self._async_job_thread)(
                query, parameters_seq
            )

    @wraps(AsyncBaseCursor._get_next_range)
    def _get_next_range(self, size: int) -> Tuple[int, int]:
        with self._idx_lock:
            return super()._get_next_range(size)

    @wraps(AsyncBaseCursor.fetchone)
    def fetchone(self) -> Optional[List[ColType]]:
        with self._query_lock.gen_rlock():
            return super().fetchone()

    @wraps(AsyncBaseCursor.fetchmany)
    def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        with self._query_lock.gen_rlock():
            return super().fetchmany(size)

    @wraps(AsyncBaseCursor.fetchall)
    def fetchall(self) -> List[List[ColType]]:
        with self._query_lock.gen_rlock():
            return super().fetchall()

    @wraps(AsyncBaseCursor.nextset)
    def nextset(self) -> None:
        with self._query_lock.gen_rlock(), self._idx_lock:
            return super().nextset()

    # Iteration support
    @check_not_closed
    @check_query_executed
    def __iter__(self) -> Generator[List[ColType], None, None]:
        while True:
            row = self.fetchone()
            if row is None:
                return
            yield row



from typing import Any, Optional

from async_property import async_cached_property  # type: ignore
from httpx import AsyncClient as HttpxAsyncClient
from httpx import Client as HttpxClient
from httpx import _types
from httpx import codes as HttpxCodes
from httpx._types import AuthTypes

from firebolt.client.auth import Auth
from firebolt.client.constants import DEFAULT_API_URL
from firebolt.common.exception import AccountNotFoundError
from firebolt.common.urls import ACCOUNT_BY_NAME_URL, ACCOUNT_URL
from firebolt.common.util import cached_property, fix_url_schema, mixin_for

FireboltClientMixinBase = mixin_for(HttpxClient)  # type: Any


class FireboltClientMixin(FireboltClientMixinBase):
    """HttpxAsyncClient mixin with Firebolt authentication functionality."""

    def __init__(
        self,
        *args: Any,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        auth: AuthTypes = None,
        **kwargs: Any,
    ):
        self.account_name = account_name
        self._api_endpoint = fix_url_schema(api_endpoint)
        super().__init__(*args, auth=auth, **kwargs)

    def _build_auth(self, auth: _types.AuthTypes) -> Optional[Auth]:
        """Create Auth objects based on auth provided.

        Overrides ``httpx.Client._build_auth``

        Args:
            auth (_types.AuthTypes): Provided auth

        Returns:
            Optional[Auth]: Auth object

        Raises:
            TypeError: Auth argument has unsupported type
        """
        if auth is None or isinstance(auth, Auth):
            return auth
        if isinstance(auth, tuple):
            return Auth(
                username=str(auth[0]),
                password=str(auth[1]),
                api_endpoint=self._api_endpoint,
            )
        raise TypeError(f'Invalid "auth" argument: {auth!r}')


class Client(FireboltClientMixin, HttpxClient):
    """An HTTP client, based on httpx.Client.

    Handles the authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple or as a
    FireboltAuth instance
    """

    @cached_property
    def account_id(self) -> str:
        """User account id.

        If account_name was provided during Client construction, returns it's id.
        Gets default account otherwise

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        if self.account_name:
            response = self.get(
                url=ACCOUNT_BY_NAME_URL, params={"account_name": self.account_name}
            )
            if response.status_code == HttpxCodes.NOT_FOUND:
                raise AccountNotFoundError(self.account_name)
            # process all other status codes
            response.raise_for_status()
            return response.json()["account_id"]

        # account_name isn't set, use the default account.
        return self.get(url=ACCOUNT_URL).json()["account"]["id"]


class AsyncClient(FireboltClientMixin, HttpxAsyncClient):
    """An HTTP client, based on httpx.AsyncClient.

    Asyncronously handles authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple or as a
    FireboltAuth instance
    """

    @async_cached_property
    async def account_id(self) -> str:
        """User account id.

        If account_name was provided during AsyncClient construction, returns it's id.
        Gets default account otherwise

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        if self.account_name:
            response = await self.get(
                url=ACCOUNT_BY_NAME_URL, params={"account_name": self.account_name}
            )
            if response.status_code == HttpxCodes.NOT_FOUND:
                raise AccountNotFoundError(self.account_name)
            # process all other status codes
            response.raise_for_status()
            return response.json()["account_id"]

        # account_name isn't set; use the default account.
        return (await self.get(url=ACCOUNT_URL)).json()["account"]["id"]



from time import time
from typing import Generator, Optional

from httpx import Auth as HttpxAuth
from httpx import Request, Response, codes

from firebolt.client.constants import _REQUEST_ERRORS, DEFAULT_API_URL
from firebolt.common.exception import AuthenticationError
from firebolt.common.token_storage import TokenSecureStorage
from firebolt.common.urls import AUTH_URL
from firebolt.common.util import cached_property, fix_url_schema


class Auth(HttpxAuth):
    """Authentication class for Firebolt database.

    Gets authentication token using
    provided credentials and updates it when it expires

    Args:
        username (str): Username
        password (str): Password
        api_endpoint (Optional[str]): Environment api endpoint.
            Default api.app.firebolt.io

    Attributes:
        username (str): Username
        password (str): Password
    """

    __slots__ = (
        "username",
        "password",
        "_api_endpoint",
        "_token",
        "_expires",
        "_use_token_cache",
    )

    requires_response_body = True

    @staticmethod
    def from_token(token: str) -> "Auth":
        """Create auth based on already acquired token.

        Args:
            token (str): Bearer token

        Returns:
            Auth: Auth object
        """
        a = Auth("", "", use_token_cache=False)
        a._token = token
        return a

    def __init__(
        self,
        username: str,
        password: str,
        api_endpoint: str = DEFAULT_API_URL,
        use_token_cache: bool = True,
    ):
        self.username = username
        self.password = password
        self._use_token_cache = use_token_cache

        # Add schema to url if it's missing
        self._api_endpoint = fix_url_schema(api_endpoint)
        self._token: Optional[str] = self._get_cached_token()
        self._expires: Optional[int] = None

    def copy(self) -> "Auth":
        """Make another auth object with same credentials.

        Returns:
            Auth: Auth object
        """
        return Auth(self.username, self.password, self._api_endpoint)

    @property
    def token(self) -> Optional[str]:
        """Acquired bearer token.

        Returns:
            Optional[str]: Acquired token
        """
        return self._token

    @property
    def expired(self) -> bool:
        """Check if current token is expired.

        Returns:
            bool: True if expired, False otherwise
        """
        return self._expires is not None and self._expires <= int(time())

    @cached_property
    def _token_storage(self) -> TokenSecureStorage:
        """Token filesystem cache storage.

        This is evaluated lazily, only if caching is enabled

        Returns:
            TokenSecureStorage: Token filesystem cache storage
        """
        return TokenSecureStorage(username=self.username, password=self.password)

    def _get_cached_token(self) -> Optional[str]:
        """If caching enabled, get token from filesystem cache.

        If caching is disabled, None is returned

        Returns:
            Optional[str]: Token if any and if caching is enabled, None otherwise
        """
        if not self._use_token_cache:
            return None
        return self._token_storage.get_cached_token()

    def _cache_token(self) -> None:
        """If caching enabled, cache token to filesystem."""
        if not self._use_token_cache:
            return
        # Only cache if token and expiration are retrieved
        if self._token and self._expires:
            self._token_storage.cache_token(self._token, self._expires)

    def get_new_token_generator(self) -> Generator[Request, Response, None]:
        """Get new token using username and password.

        Yields:
            Request: An http request to get token. Expects Response to be sent back

        Raises:
            AuthenticationError: Error while authenticating with provided credentials
        """
        try:
            response = yield Request(
                "POST",
                AUTH_URL.format(api_endpoint=self._api_endpoint),
                headers={
                    "Content-Type": "application/json;charset=UTF-8",
                    "User-Agent": "firebolt-sdk",
                },
                json={"username": self.username, "password": self.password},
            )
            response.raise_for_status()

            parsed = response.json()
            self._check_response_error(parsed)

            self._token = parsed["access_token"]
            self._expires = int(time()) + int(parsed["expires_in"])
            self._cache_token()

        except _REQUEST_ERRORS as e:
            raise AuthenticationError(repr(e), self._api_endpoint) from e

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """Add authorization token to request headers.

        Overrides ``httpx.Auth.auth_flow``

        Args:
            request (Request): Request object to update

        Yields:
            Request: Request required for auth flow
        """
        if not self.token or self.expired:
            yield from self.get_new_token_generator()

        request.headers["Authorization"] = f"Bearer {self.token}"

        response = yield request

        if response.status_code == codes.UNAUTHORIZED:
            yield from self.get_new_token_generator()
            request.headers["Authorization"] = f"Bearer {self.token}"
            yield request

    def _check_response_error(self, response: dict) -> None:
        """Check if response data contains errors.

        Args:
            response (dict): Response data

        Raises:
            AuthenticationError: Were unable to authenticate
        """
        if "error" in response:
            raise AuthenticationError(
                response.get("message", "unknown server error"),
                self._api_endpoint,
            )



from pydantic import BaseSettings, Field, SecretStr, root_validator


class Settings(BaseSettings):
    """Settings for Firebolt SDK.

    Attributes:
        user (Optional[str]): User name
        password (Optional[str]): User password
        access_token (Optional[str]): Access token to use for authentication.
            Mutually exclusive with user and password
        account_name (Optional[str]): Account name.
            Default user account is used if none provided
        server (Optional[str]): Environment api endpoint (Advanced).
            Default api endpoint is used if none provided
        default_region (str): Default region for provisioning
    """

    # Authorization
    user: str = Field(None, env="FIREBOLT_USER")
    password: SecretStr = Field(None, env="FIREBOLT_PASSWORD")
    # Or
    access_token: str = Field(None, env="FIREBOLT_AUTH_TOKEN")

    account_name: str = Field(None, env="FIREBOLT_ACCOUNT")
    server: str = Field(..., env="FIREBOLT_SERVER")
    default_region: str = Field(..., env="FIREBOLT_DEFAULT_REGION")
    use_token_cache: bool = Field(True)

    class Config:
        """Internal pydantic config."""

        env_file = ".env"

    @root_validator
    def mutual_exclusive_with_creds(cls, values: dict) -> dict:
        """Validate that either creds or token is provided.

        Args:
            values (dict): settings initial values

        Returns:
            dict: Validated settings values

        Raises:
            ValueError: Either both or none of credentials and token are provided
        """
        if values["user"] or values["password"]:
            if values["access_token"]:
                raise ValueError("Provide only one of user/password or access_token")
        elif not values["access_token"]:
            raise ValueError("Provide either user/password or access_token")
        return values



from __future__ import annotations

import socket
from json import JSONDecodeError
from types import TracebackType
from typing import Any, Callable, List, Optional, Type

from httpcore.backends.auto import AutoBackend
from httpcore.backends.base import AsyncNetworkStream
from httpx import AsyncHTTPTransport, HTTPStatusError, RequestError, Timeout
from httpx._types import AuthTypes

from firebolt.async_db.cursor import BaseCursor, Cursor
from firebolt.client import DEFAULT_API_URL, AsyncClient, Auth
from firebolt.common.exception import (
    ConfigurationError,
    ConnectionClosedError,
    FireboltEngineError,
    InterfaceError,
)
from firebolt.common.urls import (
    ACCOUNT_ENGINE_BY_NAME_URL,
    ACCOUNT_ENGINE_URL,
    ACCOUNT_ENGINE_URL_BY_DATABASE_NAME,
)
from firebolt.common.util import fix_url_schema

DEFAULT_TIMEOUT_SECONDS: int = 5
KEEPALIVE_FLAG: int = 1
KEEPIDLE_RATE: int = 60  # seconds


async def _resolve_engine_url(
    engine_name: str,
    auth: AuthTypes,
    api_endpoint: str,
    account_name: Optional[str] = None,
) -> str:
    async with AsyncClient(
        auth=auth,
        base_url=api_endpoint,
        account_name=account_name,
        api_endpoint=api_endpoint,
    ) as client:
        try:
            account_id = await client.account_id
            response = await client.get(
                url=ACCOUNT_ENGINE_BY_NAME_URL.format(account_id=account_id),
                params={"engine_name": engine_name},
            )
            response.raise_for_status()
            engine_id = response.json()["engine_id"]["engine_id"]
            response = await client.get(
                url=ACCOUNT_ENGINE_URL.format(
                    account_id=account_id, engine_id=engine_id
                ),
            )
            response.raise_for_status()
            return response.json()["engine"]["endpoint"]
        except HTTPStatusError as e:
            # Engine error would be 404.
            if e.response.status_code != 404:
                raise InterfaceError(f"Unable to retrieve engine endpoint: {e}.")
            # Once this is point is reached we've already authenticated with
            # the backend so it's safe to assume the cause of the error is
            # missing engine.
            raise FireboltEngineError(f"Firebolt engine {engine_name} does not exist.")
        except (JSONDecodeError, RequestError, RuntimeError, HTTPStatusError) as e:
            raise InterfaceError(f"Unable to retrieve engine endpoint: {e}.")


async def _get_database_default_engine_url(
    database: str,
    auth: AuthTypes,
    api_endpoint: str,
    account_name: Optional[str] = None,
) -> str:
    async with AsyncClient(
        auth=auth,
        base_url=api_endpoint,
        account_name=account_name,
        api_endpoint=api_endpoint,
    ) as client:
        try:
            account_id = await client.account_id
            response = await client.get(
                url=ACCOUNT_ENGINE_URL_BY_DATABASE_NAME.format(account_id=account_id),
                params={"database_name": database},
            )
            response.raise_for_status()
            return response.json()["engine_url"]
        except (
            JSONDecodeError,
            RequestError,
            RuntimeError,
            HTTPStatusError,
            KeyError,
        ) as e:
            raise InterfaceError(f"Unable to retrieve default engine endpoint: {e}.")


def _validate_engine_name_and_url(
    engine_name: Optional[str], engine_url: Optional[str]
) -> None:
    if engine_name and engine_url:
        raise ConfigurationError(
            "Both engine_name and engine_url are provided. Provide only one to connect."
        )


def _get_auth(
    username: Optional[str],
    password: Optional[str],
    access_token: Optional[str],
    api_endpoint: str,
    use_token_cache: bool,
) -> Auth:
    """Create Auth class based on provided credentials.

    If access_token is provided, it's used for Auth creation.
    Username and password are used otherwise.

    Returns:
        Auth: auth object

    Raises:
        ConfigurationError: Invalid combination of credentials provided

    """
    if not access_token:
        if not username or not password:
            raise ConfigurationError(
                "Neither username/password nor access_token are provided. Provide one"
                " to authenticate"
            )
        return Auth(username, password, api_endpoint, use_token_cache)
    if username or password:
        raise ConfigurationError(
            "Either username/password and access_token are provided. Provide only one"
            " to authenticate"
        )
    return Auth.from_token(access_token)


def async_connect_factory(connection_class: Type) -> Callable:
    async def connect_inner(
        database: str = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        access_token: Optional[str] = None,
        engine_name: Optional[str] = None,
        engine_url: Optional[str] = None,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        use_token_cache: bool = True,
    ) -> Connection:
        """
        Connect to Firebolt database.

        Args:
            database (str): Name of the database to connect
            username (Optional[str]): User name to use for authentication
            password (Optional[str]): Password to use for authentication
            access_token (Optional[str]): Authentication token to use insead of
                                          credentials
            engine_name (Optional[str]): The name of the engine to connect to
            engine_url (Optional[str]): The engine endpoint to use
            account_name (Optional[str]): For customers with multiple accounts;
                                          if None uses default.
            api_endpoint (str): Firebolt API endpoint. Used for authentication.
            use_token_cache (bool): Cached authentication token in filesystem.
                                    Default: True

        Note:
            Providing both `engine_name` and `engine_url` would result in an error.

        """
        # These parameters are optional in function signature
        # but are required to connect.
        # PEP 249 recommends making them kwargs.
        if not database:
            raise ConfigurationError("database name is required to connect.")

        _validate_engine_name_and_url(engine_name, engine_url)
        auth = _get_auth(
            username, password, access_token, api_endpoint, use_token_cache
        )
        api_endpoint = fix_url_schema(api_endpoint)

        # Mypy checks, this should never happen
        assert database is not None

        if not engine_name and not engine_url:
            engine_url = await _get_database_default_engine_url(
                database=database,
                auth=auth,
                account_name=account_name,
                api_endpoint=api_endpoint,
            )

        elif engine_name:
            engine_url = await _resolve_engine_url(
                engine_name=engine_name,
                auth=auth,
                account_name=account_name,
                api_endpoint=api_endpoint,
            )

        assert engine_url is not None

        engine_url = fix_url_schema(engine_url)
        return connection_class(engine_url, database, auth, api_endpoint)

    return connect_inner


class OverriddenHttpBackend(AutoBackend):
    """
    This class is a short-term solution for TCP keep-alive issue:
    https://docs.aws.amazon.com/elasticloadbalancing/latest/network/network-load-balancers.html#connection-idle-timeout
    Since httpx creates a connection right before executing a request
    backend has to be overridden in order to set the socket KEEPALIVE
    and KEEPIDLE settings.
    """

    async def connect_tcp(
        self,
        host: str,
        port: int,
        timeout: Optional[float] = None,
        local_address: Optional[str] = None,
    ) -> AsyncNetworkStream:
        stream = await super().connect_tcp(
            host, port, timeout=timeout, local_address=local_address
        )
        # Enable keepalive
        stream.get_extra_info("socket").setsockopt(
            socket.SOL_SOCKET, socket.SO_KEEPALIVE, KEEPALIVE_FLAG
        )
        # MacOS does not have TCP_KEEPIDLE
        if hasattr(socket, "TCP_KEEPIDLE"):
            keepidle = socket.TCP_KEEPIDLE
        else:
            keepidle = 0x10  # TCP_KEEPALIVE on mac

        # Set keepalive to 60 seconds
        stream.get_extra_info("socket").setsockopt(
            socket.IPPROTO_TCP, keepidle, KEEPIDLE_RATE
        )
        return stream


class BaseConnection:
    client_class: type
    cursor_class: type
    __slots__ = (
        "_client",
        "_cursors",
        "database",
        "engine_url",
        "api_endpoint",
        "_is_closed",
    )

    def __init__(
        self,
        engine_url: str,
        database: str,
        auth: AuthTypes,
        api_endpoint: str = DEFAULT_API_URL,
    ):
        # Override tcp keepalive settings for connection
        transport = AsyncHTTPTransport()
        transport._pool._network_backend = OverriddenHttpBackend()

        self._client = AsyncClient(
            auth=auth,
            base_url=engine_url,
            api_endpoint=api_endpoint,
            timeout=Timeout(DEFAULT_TIMEOUT_SECONDS, read=None),
            transport=transport,
        )
        self.api_endpoint = api_endpoint
        self.engine_url = engine_url
        self.database = database
        self._cursors: List[BaseCursor] = []
        self._is_closed = False

    def _cursor(self, **kwargs: Any) -> BaseCursor:
        """
        Create new cursor object.
        """

        if self.closed:
            raise ConnectionClosedError("Unable to create cursor: connection closed")

        c = self.cursor_class(self._client, self, **kwargs)
        self._cursors.append(c)
        return c

    async def _aclose(self) -> None:
        """Close connection and all underlying cursors."""
        if self.closed:
            return

        # self._cursors is going to be changed during closing cursors
        # after this point no cursors would be added to _cursors, only removed since
        # closing lock is held, and later connection will be marked as closed
        cursors = self._cursors[:]
        for c in cursors:
            # Here c can already be closed by another thread,
            # but it shouldn't raise an error in this case
            c.close()
        await self._client.aclose()
        self._is_closed = True

    @property
    def closed(self) -> bool:
        """True if connection is closed, False otherwise."""
        return self._is_closed

    def _remove_cursor(self, cursor: Cursor) -> None:
        # This way it's atomic
        try:
            self._cursors.remove(cursor)
        except ValueError:
            pass

    def commit(self) -> None:
        """Does nothing since Firebolt doesn't have transactions"""

        if self.closed:
            raise ConnectionClosedError("Unable to commit: connection closed")


class Connection(BaseConnection):
    """
    Firebolt asyncronous database connection class. Implements `PEP 249`_.

    Args:
        engine_url: Firebolt database engine REST API url
        database: Firebolt database name
        username: Firebolt account username
        password: Firebolt account password
        api_endpoint: Optional. Firebolt API endpoint. Used for authentication.

    Note:
        Firebolt currenly doesn't support transactions
        so commit and rollback methods are not implemented.

    .. _PEP 249:
        https://www.python.org/dev/peps/pep-0249/

    """

    cursor_class = Cursor

    aclose = BaseConnection._aclose

    def cursor(self) -> Cursor:
        c = super()._cursor()
        assert isinstance(c, Cursor)  # typecheck
        return c

    # Context manager support
    async def __aenter__(self) -> Connection:
        if self.closed:
            raise ConnectionClosedError("Connection is already closed")
        return self

    async def __aexit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        await self._aclose()


connect = async_connect_factory(Connection)
